/* UnwindProtect.h       -*-objc-*-
 *
 * Copyright Niels Mller <nisse@lysator.liu.se> 1995
 *
 * Freely distributable under the terms and conditions of the
 * GNU General Public License.
 */

#ifndef UNWINDPROTECT_H_INCLUDED
#define UNWINDPROTECT_H_INCLUDED

#include <objects/StackFrame.h>
#include <objects/Catch.h>

/* NOTE: Cleanup actions registered with cleanupByJumpingHere 
 * will *not* be called automatically if the program exit()s, nor
 * as a result of a cleanup message. */

/* UnwindProtect
 * -------------
 *
 * id tag = [UnwindProtect new];
 *
 * [tag cleanupByCalling: someFunction];
 * [tag cleanupByCalling: someFunction with: argument];
 * [tag cleanupBySending: someSelector to: someObject];
 * if (set_catch([tag cleanupByJumpingHere]))
 * {
 *    ... cleanup code ...
 *    [tag continue];
 * }
 * else
 * {
 *    ... body ...
 *
 * }
 *
 * [tag cleanup];
 *
 * Any number of cleanup actions are allowed.
 */

@interface UnwindProtect : StackFrame

/* By default, freeing this object itself is done automatically when
 * cleaning up. If you don't want this, use
 * [[UnwindProtect alloc] initDontFree] instead of [UnwindProtect new].
 */
 
- initDontFree;
- cleanupByCalling: ( void (*)(void)) function;
- cleanupByCalling: ( void (*)(void *)) function with: (void *) argument;
- cleanupBySending: (SEL) message to: rec;
- (JMP_BUF *) cleanupByJumpingHere: (frame_id *) framepointer;
- (void) cleanup;
@end

#define UNWINDPROTECT(body, cleanup_code) do {\
  frame_id _UNWIND_PROTECT_frame; \
  id _PROTECT_tag = [[UnwindProtect alloc] initDontFree]; \
  if (SETJMP(*[_PROTECT_tag \
		cleanupByJumpingHere: &_UNWIND_PROTECT_frame])==0) \
    body \
  cleanup_code \
  [[_PROTECT_tag class] unwindContinue: _UNWIND_PROTECT_frame]; \
  [_PROTECT_tag cleanup]; \
} while (0)

#endif UNWINDPROTECT_H_INCLUDED
